@@ -73,7 +73,7 @@ class SqlStorage:
         # Python-to-SQL type mapper
         self.infer_type = infer_type
 
-    def _get_writer(self, prefix):
+    def _get_writer(self, **kwargs):
         """Get a DB inserter object"""
         # This is DB-specific
         raise NotImplementedError('SqlStorage._get_writer')
@@ -239,8 +239,8 @@ class SqlStorage:
                 excluded.append(f'"{col}" = EXCLUDED."{col}"')
         return ', '.join(excluded)
 
-    def upsert(self, cursor, tablename, obj, query_id):
-        colnames = obj.keys()
+    def upsert(self, cursor, tablename, obj, query_id, schema):
+        colnames = list(schema.keys())
         excluded = self._get_excluded(colnames, tablename)
         valnames = ', '.join([f'"{x}"' for x in colnames])
         placeholders = ', '.join([self.placeholder] * len(obj))
@@ -248,28 +248,47 @@ class SqlStorage:
         if 'id' in colnames:
             stmt += f' ON CONFLICT (id) DO UPDATE SET {excluded}'
         values = tuple([str(orjson.dumps(value), 'utf-8')
-                        if isinstance(value, list) else value for value in obj.values()])
+                        if isinstance(value, list) else value for value in obj])
         logger.debug('_upsert: "%s"', stmt)
         cursor.execute(stmt, values)
 
         if query_id:
             # Now add to query table as well
+            idx = colnames.index('id')
             stmt = (f'INSERT INTO "__queries" (sco_id, query_id)'
                     f' VALUES ({self.placeholder}, {self.placeholder})')
-            cursor.execute(stmt, (obj['id'], query_id))
+            cursor.execute(stmt, (obj[idx], query_id))
 
     def upsert_many(self, cursor, tablename, objs, query_id, schema):
         for obj in objs:
-            self.upsert(cursor, tablename, obj, query_id)
+            self.upsert(cursor, tablename, obj, query_id, schema)
 
-    def cache(self, query_id, bundles, batchsize=2000):
-        """Cache the result of a query"""
+    def cache(self, query_id, bundles, batchsize=2000, **kwargs):
+        """Cache the result of a query/dataset
+
+        Takes the `observed-data` SDOs from `bundles` and "flattens"
+        them, splits out SCOs by type, and inserts into a database
+        with 1 table per type.
+
+        Accepts some keyword args for runtime options, some of which
+        may depend on what database type is in use (e.g. sqlite3,
+        postgresql, ...)
+
+        Args:
+
+          query_id (str): a unique identifier for this set of bundles
+
+          bundles (list): STIX bundles (either in-memory Python objects or filename paths)
+
+          batchsize (int): number of objects to insert in 1 batch (defaults to 2000)
+
+        """
         logger.debug('Caching %s', query_id)
 
         if not isinstance(bundles, list):
             bundles = [bundles]
 
-        writer = self._get_writer(str(query_id))
+        writer = self._get_writer(**kwargs)
         splitter = SplitWriter(writer, batchsize=batchsize, query_id=str(query_id))
 
         # walk the bundles and figure out all the columns
@@ -307,7 +326,7 @@ class SqlStorage:
                 query_id = objects[0]['query_id']
             else:
                 query_id = str(uuid.uuid4())
-        writer = self._get_writer(query_id)
+        writer = self._get_writer(query_id=query_id)
         splitter = SplitWriter(writer, batchsize=1000, query_id=str(query_id))
 
         for obj in objects:
@@ -344,12 +363,16 @@ class SqlStorage:
         if 'id' not in objects[0]:
             # Maybe it's aggregates?  Do "copy-on-write"
             self._execute(f'DROP VIEW IF EXISTS "{viewname}"', cursor)
-            columns = {k: self.infer_type(k, v) for k, v in objects[0].items()}
-            self._create_table(viewname, columns)
-            self.upsert_many(cursor, viewname, objects, None, columns)
+            columns = list(objects[0].keys())
+            schema = {}
+            for col in columns:
+                schema[col] = self.infer_type(col, objects[0][col])
+            self._create_table(viewname, schema)
+            records = [[obj.get(col) for col in columns] for obj in objects]
+            self.upsert_many(cursor, viewname, records, None, schema)
             viewdef = self._select(viewname)
         else:
-            writer = self._get_writer(None)
+            writer = self._get_writer()
             splitter = SplitWriter(writer, batchsize=1000, replace=True)
             sco_type = None
             for obj in objects:
@@ -373,7 +396,7 @@ class SqlStorage:
 
     def update(self, objects, query_id=None):
         """Update `objects`"""
-        writer = self._get_writer(None)
+        writer = self._get_writer()
         splitter = SplitWriter(writer, batchsize=1000, replace=True)
         for obj in objects:
             if 'type' not in obj:
@@ -609,3 +632,8 @@ class SqlStorage:
 
         self.connection.commit()
         cursor.close()
+
+    def finish(self):
+        """Do any DB-specific post-caching/insertion activity, such as indexing"""
+        # This is a DB-specific hook, but by default we'll do nothing
+        pass
